( function( $ ) {

	$.fn.adminMapkitLocationPicker = function() {
		return this.each( function() {
			const $mapContainer     = $( this )
			    , $map              = $( ".admin-mapkit-picker-map", $mapContainer )
			    , $postcodeField    = $( "[name="+$map.data( "postcodeField" )+"]" )
			    , $latitudeField    = $( "[name="+$map.data( "latitudeField" )+"]" )
			    , $longitudeField   = $( "[name="+$map.data( "longitudeField" )+"]" )
			    , addressFields     = $map.data( "addressFields" ).split(",")
			    , authTokenUrl      = cfrequest.mapkitData.authTokenUrl
			    , authToken         = cfrequest.mapkitData.authToken
			    , alertPosition     = cfrequest.mapkitData.alertPosition
			    , defaultLatitude   = cfrequest.mapkitData.defaultLatitude
			    , defaultLongitude  = cfrequest.mapkitData.defaultLongitude
			    , defaultDelta      = cfrequest.mapkitData.defaultDelta
			    , focusDelta        = cfrequest.mapkitData.focusDelta
			    , defaultCoord      = new mapkit.Coordinate( defaultLatitude, defaultLongitude )
			    , errorMessages     = cfrequest.mapkitData.errors
			    , combinedLatLng    = $map.data( "latitudeField" ) == $map.data( "longitudeField" )
			    , latitude          = getLatLngValues().lat
			    , longitude         = getLatLngValues().lng
			    , hasInitLatLng     = $.isNumeric( latitude ) && $.isNumeric( longitude )
			    , initLatLngCoord   = hasInitLatLng ? new mapkit.Coordinate( latitude, longitude ) : null
			    , initDelta         = hasInitLatLng ? focusDelta : defaultDelta
			    , marker            = initialiseMarker();

			let geocoder;


			// ======================
			// Initialise MapKit JS
			// ======================

			mapkit.init( {
				authorizationCallback : function( done ) {
					if ( authToken && authToken.length ) {
						done( authToken );
					} else {
						fetch( authTokenUrl )
							.then( res => res.text() )
							.then( done );
					}
				}
			} );


			// ======================
			// Initialise map
			// ======================

			const map = new mapkit.Map( $map[0], {
				  annotations : [ marker ]
				, region      : new mapkit.CoordinateRegion( marker.coordinate, new mapkit.CoordinateSpan( initDelta, initDelta ) )
			} );


			// ======================
			// Set up listeners
			// ======================

			marker.addEventListener( "drag-end", function( event ) {
				map.setCenterAnimated( marker.coordinate );
				updateFormFields();
			} );

			$latitudeField.on( "change", updateMarkerFromFields );
			$longitudeField.on( "change", updateMarkerFromFields );
			$mapContainer
				.on( "click", ".marker-set-at-postcode", lookupPostcode )
				.on( "click", ".marker-set-at-address", lookupAddress )
				.on( "click", ".marker-set-at-centre", function( event ){
					marker.coordinate = map.center;
					marker.visible    = true;
					marker.selected   = true;
					updateFormFields();
				} )
				.on( "click", ".marker-reset", function( event ){
					if ( hasInitLatLng ) {
						marker.coordinate = initLatLngCoord;
						marker.visible    = true;
						marker.selected   = true;
						map.setCenterAnimated( marker.coordinate );
					} else {
						marker.visible  = false;
						marker.selected = false;
					}
					updateFormFields();
				} );


			// ======================
			// Define functions
			// ======================

			function initialiseMarker() {
				return new mapkit.MarkerAnnotation( hasInitLatLng ? initLatLngCoord : defaultCoord, {
					  draggable : true
					, visible   : hasInitLatLng
					, selected  : hasInitLatLng
				} );
			}

			function updateMarkerFromFields() {
				let lat = getLatLngValues().lat
				  , lng = getLatLngValues().lng;

				if ( $.isNumeric( lat ) && $.isNumeric( lng ) ) {
					marker.coordinate = new mapkit.Coordinate( lat, lng );
					map.setCenterAnimated( marker.coordinate );
				}
			}

			function updateFormFields() {
				let lat = marker.visible ? marker.coordinate.latitude.toFixed( 6 )  : null
				  , lng = marker.visible ? marker.coordinate.longitude.toFixed( 6 ) : null;

				if ( combinedLatLng ) {
					if ( lat && lng ) {
						$latitudeField.val( lat + "," + lng );
					} else {
						$latitudeField.val( "" );
					}
				} else {
					$latitudeField.val( lat );
					$longitudeField.val( lng );
				}
			}

			function getLatLngValues() {
				let latLng, lat, lng;

				if ( combinedLatLng ) {
					latLng = $latitudeField.val().split( "," );
					if ( latLng.length == 2 && latLng[ 0 ].length && latLng[ 1 ].length ) {
						lat = parseFloat( latLng[ 0 ] );
						lng = parseFloat( latLng[ 1 ] );
					}
				} else {
					lat = parseFloat( $latitudeField.val()  );
					lng = parseFloat( $longitudeField.val() );
				}

				return { lat:lat, lng:lng };
			}

			function processLookupResult( type, errorMessage, result ) {
				if ( errorMessage ) {
					$.gritter.options.position = alertPosition;
					$.gritter.add( {
						  title      : errorMessages[ type ].title
						, text       : errorMessage
						, class_name : "gritter-error preside-frontend-msgbox"
						, sticky     : false
					} );
					return;
				}

				marker.coordinate = result.coordinate;
				marker.visible    = true;
				marker.selected   = true;
				map.setRegionAnimated( new mapkit.CoordinateRegion( result.coordinate, new mapkit.CoordinateSpan( focusDelta, focusDelta ) ) );
				updateFormFields();
			}

			function doLookup( type, searchString ) {
				let errorMessage, result;

				if ( !searchString.length ) {
					errorMessage = errorMessages[ type ].notSupplied;
					processLookupResult( type, errorMessage );
					return;
				}

				geocoder = geocoder || new mapkit.Geocoder();
				geocoder.lookup( searchString, function( error, data ){
					if ( error ) {
						errorMessage = errorMessages[ type ].unexpected;
					} else if ( !data.results.length ) {
						errorMessage = errorMessages[ type ].notFound;
					} else {
						result = data.results[ 0 ];
					}
					processLookupResult( type, errorMessage, result );
				} );
			}

			function lookupAddress() {
				let address = [], item;
				for( let field of addressFields ) {
					item = $( "[name="+field+"]" ).val();
					if ( item.length ) {
						address.push( item );
					}
				}
				doLookup( "address", address.join( ", " ) );
			}

			function lookupPostcode() {
				doLookup( "postcode", $postcodeField.val() );
			}
		} );

	};

	$( ".admin-mapkit-picker" ).adminMapkitLocationPicker();

} )( presideJQuery );