/**
 * @presideService
 * @singleton
 */
component {

	property name="mapkitAuthenticationService" inject="mapkitAuthenticationService";
	property name="mapkitDefaults"              inject="coldbox:setting:mapkit.defaults";
	property name="utils"                       inject="AppleServicesUtils";


// CONSTRUCTOR
	public any function init() {
		variables.SERVICE_URL        = "https://maps-api.apple.com/v1";

		variables.accessToken        = "";
		variables.accessTokenExpires = "";

		return this;
	}


// API METHODS
	public struct function geocode( required string q ) {
		var params = {
			q = arguments.q
		};

		return _apiCall( endpoint="/geocode", params=params );
	}

	public struct function reverseGeocode( required any loc ) {
		var params = {
			loc = utils.normalizeCoordinate( arguments.loc )
		};

		return _apiCall( endpoint="/reverseGeocode", params=params );
	}

	public struct function directions(
		  required any origin
		, required any destination
	) {
		var params = {
			  origin      = utils.normalizeCoordinate( arguments.origin )
			, destination = utils.normalizeCoordinate( arguments.destination )
		};

		return _apiCall( endpoint="/directions", params=params );
	}

	public struct function etas(
		  required any   origin
		, required array destinations
	) {
		var params = {
			  origin       = utils.normalizeCoordinate( arguments.origin )
			, destinations = utils.normalizeMultipleCoordinates( arguments.destinations )
		};

		return _apiCall( endpoint="/etas", params=params );
	}


// AUTHENTICATION
	public string function getAccessToken( boolean useCache=false ) {
		if ( arguments.useCache && Len( variables.accessToken ) && IsDate( variables.accessTokenExpires ) && ( variables.accessTokenExpires > Now() ) ) {
			return variables.accessToken;
		}

		var response = "";

		try {
			http url="#variables.SERVICE_URL#/token" method="get" result="response" {
				httpparam type="header" name="Authorization" value="Bearer #mapkitAuthenticationService.getJwt()#";
			}

			var responseData = IsJson( response.filecontent ) ? DeserializeJson( response.filecontent ) : response.filecontent;

			if ( response.status_code == 200 && Len( responseData.accessToken ?: "" ) ) {
				if ( arguments.useCache ) {
					variables.accessToken        = responseData.accessToken;
					variables.accessTokenExpires = DateAdd( "s", responseData.expiresInSeconds - 60, Now() );
				}

				return responseData.accessToken;
			} else {
				throw( message=responseData.error.message ?: "Unknown error obtaining Maps Access Token" );
			}
		} catch( any e ) {
			throw(
				  errorcode = response.status_code ?: 500
				, message   = e.message
			);
		}
	}


// PRIVATE METHODS
	private struct function _apiCall(
		  required string  endpoint
		,          struct  params = {}
	) {
		var requestUrl = variables.SERVICE_URL & arguments.endpoint;
		var token      = getAccessToken( useCache=true );
		var response   = "";
		var result     = { success=false };

		try {
			http url=requestUrl method="get" result="response" {
				httpparam type="header" name="Authorization" value="Bearer #token#";

				for( var param in arguments.params ) {
					if ( !IsNull( arguments.params[ param ] ) ) {
						httpparam type="url" name="#param#" value=arguments.params[ param ];
					}
				}
			}

			var responseData = IsJson( response.filecontent ) ? DeserializeJson( response.filecontent ) : response.filecontent;

			if ( response.status_code == 200 ) {
				result.success = true;
				result.result  = responseData;
			} else {
				result.error = {
					  code    = response.status_code
					, message = responseData.message ?: response.status_text
				};
			}
		} catch( any e ) {
			result.error = {
				  code    = Val( e.errorcode )
				, message = e.message
			};
		}

		return result;
	}

}