/**
 * @presideService
 * @singleton
 */
component {

	property name="encodingUtils"         inject="encodingUtils@jwtcfml";
	property name="authenticationService" inject="MapkitAuthenticationService";
	property name="utils"                 inject="AppleServicesUtils";


// CONSTRUCTOR
	public any function init() {
		variables.SERVICE_URL  = "https://snapshot.apple-mapkit.com";
		variables.WIDTH_MIN    = 50;
		variables.WIDTH_MAX    = 640;
		variables.HEIGHT_MIN   = 50;
		variables.HEIGHT_MAX   = 640;
		variables.ZOOM_MIN     = 3;
		variables.ZOOM_MAX     = 20;
		variables.MAP_SCALES   = [ 1, 2, 3 ];
		variables.MAP_TYPES    = [ "standard", "hybrid", "satellite", "mutedStandard" ];
		variables.MARKER_TYPES = [ "dot", "balloon", "large", "img" ];

		return this;
	}

	public MapkitSnapshot function newSnapshot() {
		return new model.MapkitSnapshot( this, utils );
	}

	public string function buildSnapshotUrl( required struct params, timespan expiresIn=CreateTimeSpan( 0, 0, 20, 0 ) ) {
		var settings     = $getPresideCategorySettings( "appleservices" );
		var params       = "";
		var value        = "";
		var expires      = utils.addTimeSpanToDate( arguments.expiresIn, Now() );
		var snapshotArgs = Duplicate( arguments.params );

		snapshotArgs.expires = encodingUtils.convertDateToUnixTimestamp( expires );

		for( var arg in snapshotArgs ) {
			value = snapshotArgs[ arg ];
			if ( !IsSimpleValue( value ) ) {
				value = SerializeJSON( value );
			}
			params &= "#arg#=#EncodeForURL( value )#&";
		}

		var endpoint  = "/api/v1/snapshot?#params#teamId=#settings.team_id#&keyId=#settings.key_id#";
		var signature = authenticationService.signES256( endpoint, settings.private_key );

		return "#variables.SERVICE_URL##endpoint#&signature=#signature#";
	}

	public binary function getSnapshotBinary( required string snapshotUrl ) {
		var imageRequest = "";

		http url=arguments.snapshotUrl result="imageRequest" throwonerror=true encodeurl=false;

		return imageRequest.filecontent;
	}

	public array function validationErrors( required MapkitSnapshot snapshot ) {
		var errors = [];

		var width = arguments.snapshot.getWidth();
		if ( !IsValid( "integer", width ) || width < variables.WIDTH_MIN || width > variables.WIDTH_MAX ) {
			_addError( errors, "width", width );
		}

		var height = arguments.snapshot.getHeight();
		if ( !IsValid( "integer", height ) || height < variables.HEIGHT_MIN || height > variables.HEIGHT_MAX ) {
			_addError( errors, "height", height );
		}

		var rawCenter = arguments.snapshot.getCenter();
		var center    = utils.normalizeCoordinate( rawCenter );
		var p         = ListToArray( utils.normalizeCoordinate( center ) );
		if ( center != "auto" && !_isValidLatLng( p ) ) {
			_addError( errors, "center", SerializeJson( rawCenter ) );
		}

		var zoom = arguments.snapshot.getZoom();
		if ( Len( zoom ) && ( !IsNumeric( zoom ) || zoom < variables.ZOOM_MIN || zoom > variables.ZOOM_MAX ) ) {
			_addError( errors, "zoom", zoom );
		}

		var scale = arguments.snapshot.getScale();
		if ( Len( scale ) && ( !ArrayFind( variables.MAP_SCALES, scale ) ) ) {
			_addError( errors, "scale", scale );
		}

		var span = arguments.snapshot.getSpan();
		if ( Len( span ) ) {
			if ( !_isValidLatLng( ListToArray( span ) ) ) {
				_addError( errors, "span", span );
			}
		}

		var type = arguments.snapshot.getType();
		if ( Len( type ) && !ArrayFindNoCase( variables.MAP_TYPES, type ) ) {
			_addError( errors, "type", type );
		}

		for( var annotation in arguments.snapshot.getAnnotations() ) {
			var rawPoint = annotation.point ?: "";
			var point    = utils.normalizeCoordinate( rawPoint );
			if ( point != "center" && !_isValidLatLng( ListToArray( point ) ) ) {
				_addError( errors, "annotation.point", rawPoint );
			}

			if ( Len( annotation?.glyphText ) && !ReFind( "^[A-Za-z0-9]$", annotation.glyphText ) ) {
				_addError( errors, "annotation.glyphText", annotation.glyphText );
			}

			if ( Len( annotation?.markerStyle ) && !ArrayFindNoCase( variables.MARKER_TYPES, annotation.markerStyle ) ) {
				_addError( errors, "annotation.markerStyle", annotation.markerStyle );
			}
		}

		for( var overlay in arguments.snapshot.getOverlays() ) {
			var points = overlay.points ?: [];
			if ( !ArrayLen( points ) ) {
				_addError( errors, "overlay.points", "[]" );
			} else {
				for( var rawPoint in points ) {
					var point = utils.normalizeCoordinate( rawPoint );
					if ( !_isValidLatLng( ListToArray( point ) ) ) {
						_addError( errors, "overlay.points", SerializeJSON( points ) );
						break;
					}
				}
			}

			if ( IsNumeric( overlay?.lineWidth ) && ( !IsValid( "integer", overlay.lineWidth ) || overlay.lineWidth <= 0 ) ) {
				_addError( errors, "overlay.lineWidth", overlay.lineWidth );
			}

			if ( IsArray( overlay?.lineDash ) && ArrayLen( overlay.lineDash ) ) {
				for( var i in overlay.lineDash ) {
					if ( !IsValid( "integer", i ) || i < 0 ) {
						_addError( errors, "overlay.lineDash", SerializeJSON( overlay.lineDash ) );
						break;
					}
				}
			}
		}

		return errors;
	}


// PRIVATE HELPER METHODS
	private void function _addError( required array errors, required string prop, string value="" ) {
		var error = $translateResource( "mapkit:snapshot.validate.#arguments.prop#" ) & $translateResource( uri="mapkit:snapshot.validate.actual.value", data=[ arguments.value ] );

		ArrayAppend( arguments.errors, { type=arguments.prop, message=error } );
	}

	private boolean function _isValidLatLng( required array latLng ) {
		return ArrayLen( arguments.latLng ) == 2 && _isValidLat( arguments.latLng[ 1 ] ) && _isValidLng( arguments.latLng[ 2 ] );
	}

	private boolean function _isValidLat( required any lat ) {
		return IsNumeric( arguments.lat ) && arguments.lat >= -90 && arguments.lat <= 90;
	}

	private boolean function _isValidLng( required any lng ) {
		return IsNumeric( arguments.lng ) && arguments.lng >= -180 && arguments.lng <= 180;
	}

}