/**
 * @presideService
 * @singleton
 */
component {

	property name="weatherkitAuthenticationService" inject="WeatherkitAuthenticationService";
	property name="weatherkitDefaults"              inject="coldbox:setting:weatherkit.defaults";
	property name="utils"                           inject="AppleServicesUtils";


// CONSTRUCTOR
	public any function init() {
		variables.SERVICE_URL      = "https://weatherkit.apple.com";

		variables.cachedJwt        = "";
		variables.cachedJwtExpires = "";

		return this;
	}


// API METHODS
	public any function weather(
		  required numeric  latitude
		, required numeric  longitude
		,          boolean  currentWeather   = false
		,          boolean  forecastNextHour = false
		,          boolean  forecastHourly   = false
		,          boolean  forecastDaily    = false
		,          boolean  weatherAlerts    = false
		,          datetime hourlyStart
		,          datetime hourlyEnd
		,          date     dailyStart
		,          date     dailyEnd
		,          string   countryCode
		,          string   language         = weatherkitDefaults.language
		,          string   timezone         = weatherkitDefaults.timezone
	) {
		var endpoint = "/api/v1/weather/#arguments.language#/#arguments.latitude#/#arguments.longitude#";
		var params   = { dataSets=_listDatasets( argumentCollection=arguments )	};

		if ( arguments.forecastHourly ) {
			if ( IsDate( arguments.hourlyStart ) ) {
				params.hourlyStart = DateTimeFormat( arguments.hourlyStart, "ISO" );
			}
			if ( IsDate( arguments.hourlyEnd ) ) {
				params.hourlyEnd   = DateTimeFormat( arguments.hourlyEnd  , "ISO" );
			}
		}

		if ( arguments.forecastDaily ) {
			params.timezone = arguments.timezone;
			if ( IsDate( arguments.dailyEnd ) ) {
				params.dailyStart = DateFormat( arguments.dailyStart, "yyyy-mm-dd" );
			}
			if ( IsDate( arguments.dailyEnd ) ) {
				params.dailyEnd   = DateFormat( arguments.dailyEnd  , "yyyy-mm-dd" );
			}
		}

		if ( arguments.weatherAlerts ) {
			params.countryCode = arguments.countryCode;
		}

		return _apiCall( endpoint=endpoint, params=params );
	}

	public struct function availability(
		  required numeric latitude
		, required numeric longitude
		,          string  country
	) {
		var endpoint = "/api/v1/availability/#arguments.latitude#/#arguments.longitude#";
		var params   = { country=arguments.country };

		return _apiCall( endpoint=endpoint, params=params );
	}

	public struct function attribution( string language=weatherkitDefaults.language ) {
		return _apiCall(
			  endpoint     = "/attribution/#arguments.language#"
			, authenticate = false
		);
	}


// PRIVATE METHODS
	private any function _apiCall(
		  required string  endpoint
		,          struct  params       = {}
		,          boolean authenticate = true
	) {
		var result = { success=false };

		try {
			var response     = _doHttpRequest( argumentCollection=arguments );
			var responseData = IsJson( response.filecontent ) ? DeserializeJson( response.filecontent ) : response.filecontent;

			if ( response.status_code == 200 ) {
				result.success = true;
				result.result  = responseData;
			} else {
				result.error   = {
					  code    = response.status_code
					, message = response.status_text
					, detail  = responseData.reason ?: ""
				};
			}
		} catch( any e ) {
			result.error   = {
				  code    = 500
				, message = e.message
				, detail  = e.detail
			};
		}

		return result;
	}

	private any function _doHttpRequest(
		  required string  endpoint
		,          struct  params       = {}
		,          boolean authenticate = true
	) {
		var requestUrl = variables.SERVICE_URL & arguments.endpoint;
		var token      = _getCachedJwt();
		var response   = "";

		http url=requestUrl method="get" result="response" {
			if ( arguments.authenticate ) {
				httpparam type="header" name="Authorization" value="Bearer #token#";
			}
			for( var param in arguments.params ) {
				if ( !IsNull( arguments.params[ param ] ) ) {
					httpparam type="url" name="#param#" value=arguments.params[ param ];
				}
			}
		}

		return response;
	}

	private string function _getCachedJwt() {
		if ( Len( variables.cachedJwt ) && IsDate( variables.cachedJwtExpires ) && ( variables.cachedJwtExpires > Now() ) ) {
			return variables.cachedJwt;
		}

		variables.cachedJwt        = weatherkitAuthenticationService.getJwt( CreateTimeSpan( 0, 1, 0, 0 ) );
		variables.cachedJwtExpires = DateAdd( "n", 59, Now() );

		return variables.cachedJwt;
	}

	private string function _listDatasets() {
		var availableDatasets = [ "currentWeather", "forecastDaily", "forecastHourly", "forecastNextHour", "weatherAlerts" ];
		var datasets          = [];

		for( var dataSet in availableDatasets ) {
			if ( $helpers.isTrue( arguments[ dataset ] ?: "" ) ) {
				ArrayAppend( datasets, dataset );
			}
		}

		return ArrayToList( datasets );
	}

}